--/run print(string.format("Unhittable at 102.4%% - you have %.2f%%", GetDodgeChance() + GetBlockChance() + GetParryChance() + 5 + 1/(0.0625 + 0.956/(GetCombatRating(CR_DEFENSE_SKILL)/4.91850*0.04))))

local Unhittable = LibStub("AceAddon-3.0"):NewAddon("Broker_Unhittable", "AceEvent-3.0", "AceBucket-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("Unhittable")

local pLevel = UnitLevel("player")
local getMissChance, getDamageReduction, getResilienceEffect, getHexColor
local db

Unhittable.ldb = LibStub:GetLibrary("LibDataBroker-1.1"):NewDataObject("Broker_Unhittable", {
	type = "data source",
	icon = "Interface\\Icons\\Spell_Holy_DevotionAura",
	label = "Unhittable",
	text = "",
})

local defaults = {
	profile = {
		avoidColor = {0, 1, 0, 1},
		drColor = {0, .5, 1, 1},
		armorColor = {1, 1, 1, 1},
		headerColor = {0, .5, 1, 1},
		labelColor = {0, 1, 0, 1},
		valueColor = {1, 1, 1, 1},
		brokerLabel = 0,
		showFuBarAV = true,
		showFuBarAVBlock = false,
		showFuBarDR = true,
		showFuBarArmor = false,
		displayPrecisionFuBar = 2,
		showTooltipDefense = true,
		showTooltipCombat = true,
		showTooltipArmor = true,
		showTooltipDR = true,
		displayPrecisionTooltip = 2,
		baselineLevel = 0,
	}
}

local function getOptions()
	local options = {
		name = "Unhittable",
		type = "group",
		childGroups = "tab",
		get = function(info)
			return db[info[#info]]
		end,
		set = function(info, value)
			db[info[#info]] = value
			Unhittable:UpdateDisplay()
		end,
		args = {
			desc = {
				order = 1,
				type = "description",
				name = L["Displays a variety of useful stats for tanking."].."\n",
			},
			baselineLevel = {
				type = "range",
				name = L["Target Level (0 = Player Level)"],
				desc = L["Used for level difference in calculations."],
				min = 0,
				max = 83,
				step = 1,
				width = "full",
				order = 5,
			},
			display = {
				type = "group",
				name = L["Display"],
				desc = L["Control how values are to displayed."],
				order = 1,
				args = {
					header1 = { name = L["Data Feed"], type = "header", order = 10 },
					displayPrecisionFuBar = {
						type = "range",
						name = L["Display Precision"],
						desc = L["Number of decimal places displayed on the data feed."],
						min = 0, max = 4, step = 1,
						order = 11,
					},
					brokerLabel = {
						type = "select",
						name = L["Label Style"],
						desc = L["Style of labels to display on the data feed."],
						values = { [0] = L["None"], [1] = L["Short"], [2] = L["Long"] },
						order = 12,
					},
					showFuBarAV = {
						type = "toggle",
						name = L["Show Avoidance"],
						desc = L["Display the avoidance percentage on the data feed."],
						--width = "full",
						order = 13,
					},
					showFuBarAVBlock = {
						type = "toggle",
						name = L["Include Block with Avoidance"],
						desc = L["Include block chance in the avoidance percentage on the data feed."],
						--width = "full",
						disabled = function() return not db.showFuBarAV end,
						order = 14,
					
					},
					showFuBarDR = {
						type = "toggle",
						name = L["Show Damage Reduction"],
						desc = L["Display the damage reduction percentage on the data feed."],
						--width = "full",
						order = 15,
					},
					showFuBarArmor = {
						type = "toggle",
						name = L["Show Armor"],
						desc = L["Display the armor value on the data feed."],
						--width = "full",
						order = 16,
					},
					
					header2 = { name = L["Tooltip"], type = "header", order = 20 },
					displayPrecisionTooltip = {
						type = "range",
						name = L["Display Precision"],
						desc = L["Number of decimal places displayed in the tooltip."],
						min = 0, max = 4, step = 1,
						order = 21,
					},
					showTooltipDefense = {
						type = "toggle",
						name = L["Show Defense"],
						desc = L["Display the defense values."],
						--width = "full",
						order = 22,
					},
					showTooltipCombat = {
						type = "toggle",
						name = L["Show Combat Table"],
						desc = L["Display the combat table values."],
						--width = "full",
						order = 23,
					},
					showTooltipArmor = {
						type = "toggle",
						name = L["Show Armor"],
						desc = L["Display the armor values."],
						--width = "full",
						order = 24,
					},
				},
			},
			color = {
				type = "group",
				name = L["Colors"],
				desc = L["Set the text colors."],
				order = 2,
				get = function(info)
					return unpack(db[info[#info]])
				end,
				set = function(info, r, g, b)
					db[info[#info]] = {r, g, b, 1}
					Unhittable:UpdateDisplay()
				end,
				args = {
					brokerHeader = {
						type = "header",
						name = L["Data Feed"],
						order = 10,
					},
					avoidColor = {
						type = "color",
						name = L["Avoidance"],
						desc = L["Set the color of the avoidance percentage on the data feed."],
						order = 11,
					},
					drColor = {
						type = "color",
						name = L["Damage Reduction"],
						desc = L["Set the color of the damage reduction percentage on the data feed."],
						order = 12,
					},
					armorColor = {
						type = "color",
						name = L["Armor"],
						desc = L["Set the color of the armor value on the data feed."],
						order = 13,
					},
					tooltipHeader = {
						type = "header",
						name = L["Tooltip"],
						order = 20,
					},
					headerColor = {
						type = "color",
						name = L["Header"],
						desc = L["Set the color of header text."],
						order = 21,
					},
					labelColor = {
						type = "color",
						name = L["Label"],
						desc = L["Set the color of label text."],
						order = 22,
					},
					valueColor = {
						type = "color",
						name = L["Value"],
						desc = L["Set the color of value text."],
						order = 23,
					},
				},
			},
		}
	}
	options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(Unhittable.db)
	options.args.profile.order = 100
	return options
end

function Unhittable:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("UnhittableDB", defaults)
	db = self.db.profile
	
	LibStub("AceConfig-3.0"):RegisterOptionsTable("Broker Unhittable", getOptions)
	LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Broker Unhittable", "Broker Unhittable")
end

function Unhittable:OnEnable()
	self:RegisterEvent("PLAYER_ENTERING_WORLD", "UpdateDisplay")
	self:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED", "UpdateDisplay")
	self:RegisterEvent("CHARACTER_POINTS_CHANGED", "UpdateDisplay")
	self:RegisterEvent("PLAYER_LEVEL_UP", "UpdateDisplay")
	self:RegisterEvent("PLAYER_TALENT_UPDATE", "UpdateDisplay")
	self:RegisterBucketEvent("UNIT_INVENTORY_CHANGED", .2, "BucketUpdate")
	self:RegisterBucketEvent("UNIT_AURA", .2, "BucketUpdate")
end

function Unhittable:BucketUpdate(units)
	if units.player then
		self:UpdateDisplay()
	end
end

do
	local k = select(2, UnitClass("player")) == "DRUID" and 0.972 or 0.956
	function getMissChance()
		--1/Ad = 1/c + k/A
		--Where 
		--A is the amount of avoidance before diminishing returns
		--c is the cap for the avoidance stat
		--k is the constant for your class
		--Ad is the amount of avoidance after diminishing returns are applied (how much your avoidance actually increases)
		
		--your base miss is 5%,
		--c = 16 for miss (1/16 = 0.0625),
		--k = 0.956 for war/dk/pld, 0.972 for druid
		--and each point of defense skill increases your chance to be missed by 0.04%
		
		return 5 + 1/(0.0625 + k/(GetCombatRatingBonus(CR_DEFENSE_SKILL)*0.04))
	end
end
	
function getDamageReduction(level, base)
	if not level then level = pLevel end
	local baseArmor, effectiveArmor = UnitArmor("player")
	local armor = base and baseArmor or effectiveArmor

	local levelModifier = level
	if levelModifier > 59 then
		levelModifier = levelModifier + (4.5 * (levelModifier - 59))
	end
	local temp = armor / (85 * levelModifier + 400)
	local armorReduction = temp / (1 + temp)

	armorReduction = max(0, min(.75, armorReduction))
	armorReduction = armorReduction * 100

	return tonumber(("%.4f"):format(armorReduction))
end

function getResilienceEffect(level)
	local rating = GetCombatRating(CR_CRIT_TAKEN_MELEE)
	if rating == 0 then return rating end
	
	if not level then level = pLevel end
	if level < 34 then level = 34 end
	
	if level >= 70 then
		return rating/25/((82/52)*(131/63)^((level-70)/10))
	elseif level >= 60 then
		return rating/25/(82/(262-3*level))
	elseif level >= 10 then
		return rating/25/((level-8)/52)
	else
		return rating/25/(2/52)
	end
end

function getHexColor(color)
	--if not color[1] or not color[2] or not color[3] then return "ffffff" end
	return ("%02x%02x%02x"):format(color[1] * 255, color[2] * 255, color[3] * 255)
end


function Unhittable:UpdateDisplay()
	local calcLevel = db.baselineLevel > 0 and db.baselineLevel or pLevel
	local displayString = "|r"
	local formatString = ("%%s|cff%%s%%.%df%%%%|r "):format(db.displayPrecisionFuBar)
	
	if db.showFuBarAV then
		local defenseDiff = (calcLevel - pLevel) * 0.20
		local av = getMissChance() + GetDodgeChance() + GetParryChance() - (defenseDiff * 3)
		local avb = getMissChance() + GetDodgeChance() + GetParryChance() + GetBlockChance() - (defenseDiff * 4)
		if db.brokerLabel == 1 then
			displayString = ("%s%s"):format(displayString, L["Av:"])
		elseif db.brokerLabel == 2 then
			displayString = ("%s%s"):format(displayString, L["Avoidance:"])
		end
		displayString = (formatString):format(displayString, getHexColor(db.avoidColor), db.showFuBarAVBlock and avb or av)
	end
	
	if db.showFuBarDR then
		if db.brokerLabel == 1 then
			displayString = ("%s%s"):format(displayString, L["DR:"])
		elseif db.brokerLabel == 2 then
			displayString = ("%s%s"):format(displayString, L["Damage Reduction:"])
		end
		displayString = (formatString):format(displayString, getHexColor(db.drColor), getDamageReduction(calcLevel))
	end
	
	if db.showFuBarArmor then
		local _, totalArmor = UnitArmor("player")
		if db.brokerLabel == 1 then
			displayString = ("%s%s"):format(displayString, L["Ar:"])
		elseif db.brokerLabel == 2 then
			displayString = ("%s%s"):format(displayString, L["Armor:"])
		end
		displayString = ("%s|cff%s%d|r "):format(displayString, getHexColor(db.armorColor), totalArmor)
	end  
	
	self.ldb.text = displayString
end

--ldb functions
function Unhittable.ldb.OnClick(_, button)
	if button == "RightButton" then
		InterfaceOptionsFrame_OpenToCategory("Broker Unhittable")
	elseif button == "LeftButton" and db.showFuBarAV then
		db.showFuBarAVBlock = not db.showFuBarAVBlock
		Unhittable:UpdateDisplay()
	end
end

function Unhittable.ldb.OnTooltipShow(tooltip)
	if not tooltip or not tooltip.AddLine or not tooltip.AddDoubleLine then return end
	
	local calcLevel = db.baselineLevel > 0 and db.baselineLevel or pLevel
	local defenseDiff = (calcLevel - pLevel) * 0.20
	
	local baseArmor, effectiveArmor, _, armorBuffs, armorDebuffs = UnitArmor("player")

	local defenseRating = GetCombatRating(CR_DEFENSE_SKILL)
	
	local baseDefense, armorDefense = UnitDefense("player")
	local totalDefense = baseDefense + armorDefense
	local defenseEffect = (totalDefense - calcLevel * 5) * 0.04

	local missChance = getMissChance() - defenseDiff
	local dodgeChance = GetDodgeChance() - defenseDiff
	local parryChance = GetParryChance() - defenseDiff
	local blockChance = GetBlockChance() - defenseDiff --different at -30
	local critChance = max(0, 5 - (defenseEffect - getResilienceEffect()))
	local crushChance = 0
	if (calcLevel - pLevel) > 3 then
		crushChance = 15 + max(0, (pLevel * 5 - totalDefense) * 0.02)
	end
	local hitChance = max(0, 100 - (missChance + dodgeChance + parryChance + blockChance + critChance + crushChance))

	local formatPercentString = ("%%.%df%%%%"):format(db.displayPrecisionTooltip)
	local hr, hg, hb = unpack(db.headerColor)
	local lr, lg, lb = unpack(db.labelColor)
	local vr, vg, vb = unpack(db.valueColor)
	
	tooltip:AddLine(_G["HIGHLIGHT_FONT_COLOR_CODE"] .. "Unhittable")
	tooltip:AddLine(" ")

	if db.showTooltipDefense then
		tooltip:AddLine(L["Defense Calculation"], hr, hg, hb)
		if baseDefense == totalDefense then
			tooltip:AddDoubleLine(L["Total Defense"], ("%d"):format(totalDefense), lr, lg, lb, vr, vg, vb)
		else
			tooltip:AddDoubleLine(L["Total Defense"], ("%d (%d+%d)"):format(totalDefense, baseDefense, armorDefense), lr, lg, lb, vr, vg, vb)
		end
		if pLevel == 80 and calcLevel >= 80 then
			--just wanted a way to see how much i can play with defense rating. 1 skill = 4.91850 rating
			tooltip:AddDoubleLine(L["Defense Rating"], ("%d (%+d)"):format(defenseRating, ceil(-121.278*(5+defenseDiff-(GetDodgeBlockParryChanceFromDefense())))), lr, lg, lb, vr, vg, vb)
		else
			tooltip:AddDoubleLine(L["Defense Rating"], ("%d"):format(defenseRating), lr, lg, lb, vr, vg, vb)
		end
		--tooltip:AddDoubleLine(L["Avoidance"], (formatPercentString):format((missChance + dodgeChance + parryChance)), lr, lg, lb, vr, vg, vb)
		tooltip:AddLine(" ")
	end

	if db.showTooltipCombat then
		tooltip:AddLine(L["Combat Table"], hr, hg, hb)
		tooltip:AddDoubleLine(L["Miss Chance"], (formatPercentString):format(missChance), lr, lg, lb, vr, vg, vb)
		tooltip:AddDoubleLine(L["Dodge Chance"], (formatPercentString):format(dodgeChance), lr, lg, lb, vr, vg, vb)
		tooltip:AddDoubleLine(L["Parry Chance"], (formatPercentString):format(parryChance), lr, lg, lb, vr, vg, vb)
		tooltip:AddDoubleLine(L["Block Chance"], (formatPercentString):format(blockChance), lr, lg, lb, vr, vg, vb)
		if critChance > 0 then
			tooltip:AddDoubleLine(L["Crit Chance"], (formatPercentString):format(critChance), lr, lg, lb, vr, vg, vb)
		end
		if crushChance > 0 then
			tooltip:AddDoubleLine(L["Crush Chance"], (formatPercentString):format(crushChance), lr, lg, lb, vr, vg, vb)
		end
		tooltip:AddDoubleLine(L["Hit Chance"], (formatPercentString):format(hitChance), lr, lg, lb, vr, vg, vb)
		tooltip:AddLine(" ")
	end

	if db.showTooltipArmor then
		tooltip:AddLine(L["Armor"], hr, hg, hb)
		if baseArmor == effectiveArmor then
			tooltip:AddDoubleLine(L["Armor"], ("%d"):format(effectiveArmor), lr, lg, lb, vr, vg, vb)
		else
			tooltip:AddDoubleLine(L["Base Armor"], ("%d"):format(baseArmor), lr, lg, lb, vr, vg, vb)
			if armorBuffs > 0 then
				tooltip:AddDoubleLine(L["Armor Buffs"], ("+%d"):format(armorBuffs), lr, lg, lb, vr, vg, vb)
			end
			if armorDebuffs < 0 then
				tooltip:AddDoubleLine(L["Armor Debuffs"], ("%d"):format(armorDebuffs), lr, lg, lb, vr, vg, vb)
			end
			tooltip:AddDoubleLine(L["Total Armor"], ("%d"):format(effectiveArmor), lr, lg, lb, vr, vg, vb)
		end
		tooltip:AddDoubleLine(L["Damage Reduction"], (formatPercentString):format(getDamageReduction(calcLevel)), lr, lg, lb, vr, vg, vb)
		tooltip:AddLine(" ")
	end
end